module HParse(typecheck,typetheo,TrafoApplyMap,applytrafo,
              parse,ptheory,parseExp,parse',typecheck',SourceFileInfo (..))
where

import UU_Scanner
import UU_Parsing
import UU_Parsing_ext
import Char
import Types
import List
import Utils
import Syntax
import Infer
import Laws
import Constants
import Expressions
import SCC
----------------------------------------------------------------------------
-- type checker

data SourceFileInfo 
  = SourceFileInfo {
                    sfienv :: Env,
                    sfifile :: SourceFile,
                    sfitext :: String,
                    sfiprog :: Prog,
                    sfitransforms :: [String],
                    sfitheormap :: TheorMap,
                    sfitrafomap :: TrafoApplyMap
                   }                   

typecheck :: SourceFile -> IO SourceFileInfo
typecheck f = do let (res,errs) = typecheck' f
                 putStr (concat errs)
                 return res

typecheck' :: SourceFile -> (SourceFileInfo,[String])
typecheck' f = (SourceFileInfo {
                 sfifile = f,
                 sfienv = env,
                 sfitext = error "sourcefile text not inserted",
                 sfiprog = prog',
                 sfitransforms = trafos, 
                 sfitheormap = theormap,
                 sfitrafomap = trafomap},
                errs)
        where (trafos,strafomap,depnodes) = tcSourceFile f basicdefns
              prog = bundle depnodes
              (errs,env,theormap,prog') = tcProg prog
              trafomap name = fmap (\f -> f (env,theormap)) $ strafomap name


bundle depnodes = Prog $ foldr blockify Empty $ depordering depnodes
           where blockify [(TrafoElem t,n)] b = TrafoBlock t b
                 blockify es b 
                    = DefsBlock (DList $ foldr onedef NoDef $ map getdef es) b
                 getdef (DefElem d,n) = (d,n)
                 getdef (TrafoElem t,n) = error mrtrafoerr 
                 onedef (def,lineno) defs = OneDef def defs lineno

basicdefns = [(["(++)"],parseCorrectDef "[] ++ ys = ys"),
              (["(++)"],parseCorrectDef "(x:xs) ++ ys = x:(xs++ys)"),
              (["(.)"], parseCorrectDef "(f.g) x = f (g x)")]
   where parseCorrectDef x = let (res,_,_) = invokeErr pDef (scan x) [] in res


mrtrafoerr 
   = "Can't have mutually recursive elements involving a transformation block"

tcSourceFile (SourceFile es) = sem_SourceFile_SourceFile (tcElems es)
tcElems NoElem = sem_Elems_NoElem
tcElems (OneElem e es n) = sem_Elems_OneElem (tcElem e) (tcElems es) n
tcElem (DefElem d) = sem_Elem_DefElem (tcDef d)
tcElem (TrafoElem t) = sem_Elem_TrafoElem (tcTransform t)

tcProg (Prog b) = sem_Prog_Prog (tcBlock b)
tcBlock Empty = sem_Block_Empty
tcBlock (DefsBlock d b) = sem_Block_DefsBlock (tcDList d) (tcBlock b)
tcBlock (TrafoBlock t b) = sem_Block_TrafoBlock (tcTransform t) (tcBlock b)

tcTransform (Rewrite name depends redef spec using) 
      = sem_Transform_Rewrite name depends 
                              (tcDList redef)
                              (tcDList spec)
                              (tcTheory using)

tcDList (DList ds) = sem_DList_DList (tcDefs ds)
tcDefs  NoDef = sem_Defs_NoDef
tcDefs (OneDef d ds n) = sem_Defs_OneDef (tcDef d) (tcDefs ds) n
tcDef (Fun lhs rhs) = sem_Def_Fun (tcExp lhs) (tcExp rhs)
tcDef (Dec s t) = sem_Def_Dec (tcExp s) t 
tcDef (Data t as) = sem_Def_Data t (tcAlts as)
tcAlts NoAlt = sem_Alts_NoAlt
tcAlts (OneAlt a as) = sem_Alts_OneAlt (tcAlt a) (tcAlts as)
tcAlt (Alt n t) = sem_Alt_Alt n t 
tcExp (Lam pat body t) = sem_Exp_Lam (tcExp pat) (tcExp body) t
tcExp (Let ds body t) = sem_Exp_Let (tcDList ds) (tcExp body) t
tcExp (Ap fun arg t) = sem_Exp_Ap (tcExp fun) (tcExp arg) t
tcExp (Var name t) = sem_Exp_Var name t
tcExp (Con value t) = sem_Exp_Con value t

typetheo env th = do putStr (concat errs)
                     return theor
		       where (errs,_,_,_,theor,_) = tcTheory th env undefined

tcTheory (Theory t) = sem_Theory_Theory (tcTheo t)
tcTheo NoTheo = sem_Theo_NoTheo
tcTheo (CeqTheo c t) = sem_Theo_CeqTheo (tcCeq c) (tcTheo t)
tcTheo (DefnTheo n t) = sem_Theo_DefnTheo n (tcTheo t)

typeCeq env c = tcCeq c env
tcCeq (Ceq n c e l) = sem_Ceq_Ceq n (tcEqs c) (tcEqn e) l
tcEqs NoEq = sem_Eqs_NoEq
tcEqs (OneEq e es) = sem_Eqs_OneEq (tcEqn e) (tcEqs es)
tcEqn (Eqn l r) = sem_Eqn_Eqn (tcExp l) (tcExp r) 

lamP pat body = Lam pat body undefType
letP ds body = Let ds body undefType
apP fun arg = Ap fun arg undefType
varP name = Var name undefType
conP value = Con value undefType

----------------------------------------------------------------------------
-- transformations

type TrafoApplyMap = String -> Maybe TrafoApply

applytrafo :: Monad m => TrafoApply -> (Theor -> (String,Type) -> m [Def]) 
                                    -> m SourceFile
applytrafo (RewriteApply rewriteinfo sourcefileinfo proginfo)
   = dorewritetrafo rewriteinfo sourcefileinfo proginfo

dorewritetrafo :: Monad m => (String,[String],DList,[String]) 
                        -> [SourceElem] -> (Env,TheorMap)
                        -> (Theor -> (String,Type) -> m [Def]) -> m SourceFile

dorewritetrafo (trafo_name,redef_names,redef_dlist,spec_names) 
               sourceelems (env,theormap) calcfun
 = do spec_newdefs <- sequence $ map (calcfun th . makeexp) spec_names
      let redef_defs = defslist redef_dlist
      let newdefs = redef_defs ++ concat spec_newdefs
      return $ SourceFile 
             $ replaceelems (redef_names,[trafo_name]) 
                     (map DefElem newdefs) sourceelems
     where Just th = theormap trafo_name
           makeexp name = let Just typ = lookup name env 
                          in (name,inst (take (nbound typ) namesupply) typ)
           defslist (DList defs) = defslist' defs
           defslist' NoDef = []
           defslist' (OneDef def defs _) = def:defslist' defs
           replaceelems (redef_names,trafo_names) newelems [] 
               = error "couldn't find original definition to remove"
           replaceelems r@(redef_names,_) newelems (DefSE def_names def:elems)
               = if def_names `subseteq` redef_names 
                  then foldr oneelem (deleteelems r elems) newelems
                 else if not (null (def_names `intersect` redef_names)) 
                  then error "must redefine everything"
                 else oneelem (DefElem def) (replaceelems r newelems elems)
           replaceelems r@(_,trafo_names) newelems 
                                              (TrafoSE trafo_name trafo:elems)
               = if trafo_name `elem` trafo_names 
                  then replaceelems r newelems elems
                 else oneelem (trafoelem trafo) (replaceelems r newelems elems)
           deleteelems _ [] = NoElem
           deleteelems r@(redef_names,_) (DefSE def_names def:elems)
               = if def_names `subseteq` redef_names 
                  then deleteelems r elems
                 else if not (null (def_names `intersect` redef_names)) 
                  then error "must redefine everything"
                 else oneelem (DefElem def) (deleteelems r elems)
           deleteelems r@(_,trafo_names) (TrafoSE trafo_name trafo:elems)
               = if trafo_name `elem` trafo_names then deleteelems r elems
                 else oneelem (trafoelem trafo) (deleteelems r elems)
           oneelem elem elems = OneElem elem elems 0
           trafoelem (Rewrite name depends redef spec using) 
               = TrafoElem 
                   (Rewrite name (depends \\ [trafo_name]) redef spec using)

----------------------------------------------------------------------------
-- scanner 

keywordstxt = ["if","then","else","let","in","mod","True","False","where","data","TRANSFORM","REDEFINE","SPECIFYING","USING","END","DEFINITION"]
keywordsops = ["->"]
specialchars = ",;[](){}\\"
opchars = ".:!#$%&*+/<=>?@^|-\\" 

scan x = scanner keywordstxt keywordsops specialchars opchars x 1


----------------------------------------------------------------------------
-- parser

parse :: String -> IO SourceFile
parse x = do putStr cfree_errors
             return res
          where (res,rest,cfree_errors) = invokeErr pSourceFile (scan x) []

parse' :: String -> (SourceFile,String)
parse' x = (res,errors)
           where (res,rest,errors) = invokeErr pSourceFile (scan x) []


ptheory :: String -> IO Theory
ptheory x = do putStr cfree_errors
               return res
            where (res,rest,cfree_errors) = invokeErr pTheory (scan x) []

parseExp :: String -> IO Exp
parseExp x = do putStr cfree_errors
                return res
             where (res,rest,cfree_errors) = invokeErr pExp (scan x) []

-- definitions

{-
pProg = Prog <$> pBlock

pBlock = foldr OneBlock Empty <$> pList pDList
-}

pDlist = DList <$> pDefs pcurly_psemics_lines

pSourceFile = SourceFile <$> pElems psemics_lines

psemics_lines p = pList ((,) <$> p <*> pSymLine ';')

pElems p = foldr oneElem NoElem <$> (p pElem <|> wrap <$> pElem)
                                  <?> (NoElem, "list of source elements")
             where oneElem (e,n) es = OneElem e es (n+1)
                   wrap d = [(d,-1)]

pElem = DefElem <$> pDef <|> TrafoElem <$> pTransform

pDefs p = foldr oneDef NoDef <$> (p pDef <|> wrap <$> pDef)
                                  <?> (NoDef, "list of definitions")
              where oneDef (a,n) ds = OneDef a ds (n+1)
                    wrap d = [(d,-1)]

pTransform = pRewrite

pRewrite = Rewrite <$-> pKey "TRANSFORM"  <*> pVarid
                   <*>  (pOper ":" <-*> pcommas pVarid <|> succeed [])
                   <*-> pKey "REDEFINE"   <*> pDefBlock
                   <*-> pKey "SPECIFYING" <*> pDefBlock
                   <*-> pKey "USING"      <*> pTheory
                   <*-> pKey "END"
         where pDefBlock = DList <$> pDefs (psemics . ((\d -> (d,-1)) <$>))

pDef = pFunOrDec <|> pData 
       <?> (Fun undefExp undefExp, "definition")

pFunOrDec = (\e -> either (Fun e) (Dec e)) <$> pExp <*> por
            where por = Left <$-> pOper "=" <*> pExp <|>
                        Right <$-> pOper "::" <*> pType
				 

pData = Data <$-> pKey "data" <*> pScheme <*-> pOper "=" <*> pAlts
        <?> (Data undefType NoAlt, "data declaration")

pAlts = (foldr OneAlt NoAlt) <$> pListSep (pOper "|") pAlt 
        <?> (NoAlt, "constructor declaration(s)")

pAlt = Alt <$> pConid <*> pList pAtomScheme

-- types

pType = close  <$> pArrows 
        <?> (undefType, "type expresssion")

pArrows =  arrows <$> pListSep (pKey "->") pScheme

pAtomScheme = Tid  <$> pVarid  <|>
              pTupType <|>
              (TCon "List" . (:[]))<$> pbracks pArrows <|>
              (\ c -> TCon c []) <$> pConid

pScheme = TCon <$> pConid <*> pList pScheme   	<|>
          pAtomScheme
          
pTupType = tuple <$> pparens_pcommas pArrows
           where tuple [a] = a
                 tuple x = TCon ("Tuple"++show (length x))  x

-- expressions with where clauses

pExp = wher <$> pWexp <*> (pWhere <|> succeed (DList NoDef))
       <?> (undefExp, "expression")
       where wher b (DList NoDef) = b
             wher b ds = letP ds b

pWhere = pKey "where" <-*> pDlist

pWexp = pApp <|> pIf <|> pLet 
        <?> (undefExp, "application, conditional or local definition")

-- sequences of boxes, separated by operators

pApp    = operatorParse 	        <$> pBoxes
pBoxes  = (\b (bs,ops) -> (b:bs,ops)) 	<$> pBox <*> pOBoxes
pOBoxes = (\o (bs,ops) -> (bs,o:ops)) 	<$> pOperator <*> pBoxes 	<|> 
          (\ (bs,ops) -> (bs,App:ops)) 	<$> pBoxes 			<|> 
          succeed ([],[])

pBox = pPat   		        <|>
       pAbs	
       <?> (undefExp, "(atomic) expression")

pPat = varP <$> pVarid 		<|> 
       conP <$> pConstant 	<|> 
       pTuple 		        <|>
       pEnum	
       <?> (undefExp, "variable, constant, tuple or list enumeration")

-- abstractions and let expressions

pAbs = lam <$-> pSym '\\' <*> pSome pPat <*-> pKey "->" <*> pExp
       <?> (lamP undefExp undefExp, "abstraction")
       where lam xs e = foldr lamP e xs

pLet = letP <$-> pKey "let" <*> pDlist <*-> pKey "in" <*> pExp  
       <?> (letP (DList NoDef) undefExp, "local definition")

-- interface with constant parser:


pIf = cond <$-> pKey "if" <*> pExp <*-> pKey "then" <*> pExp <*-> pKey "else" <*> pExp
      <?> (cond undefExp undefExp undefExp, "if-then-else")
      where cond c t e= foldl1 apP [conP If, c, t, e]

pEnum = foldr cons nul <$> pbracks_pcommas_e pExp
        <?> (nul, "list enumeration")
        where cons a b = foldl1 apP [conP (Binop Cons),a,b]
              nul = conP Nul

pTuple = tuple <$> pparens (pcommas pExp <|> (((:[]). conP . (Binop)) <$> pOperator))
         <?> (undefExp, "bracketed expression or tuple")
         where tuple [a] = a
               tuple x = foldl1 apP (conP (Tuple (length x)) : x)

operatorParse (vals,ops) = curryOps (decWith vals (assoc (pparse ops)))

curryOps :: Expr Exp Op -> Exp
curryOps (Tip e) = e
curryOps (Node e1 App e2) = apP (curryOps e1) (curryOps e2)
curryOps (Node e1 op e2) = foldl1 apP [conP (Binop op), curryOps e1, curryOps e2]

undefExp = conP Error 
undefType = errType

-- equations and theories

pTheory = (Theory . foldr onetheo NoTheo) 
                <$> psemics (Left <$> pCEqn <|> Right <$> pDefnRule)
          <?> (Theory NoTheo, "theory")
     where onetheo (Left ceq) theo = CeqTheo ceq theo
           onetheo (Right names) theo = foldr DefnTheo theo names

pDefnRule = pKey "DEFINITION" <-*> pcommas (pVarid <|> pMaybeParens (pCat <|> pDot))
     where pCat = const "(++)" <$> pOper "++"
           pDot = const "(.)" <$> pOper "."
           pMaybeParens p = p <|> pparens p

pCEqn = ceq <$> pVarid <*> pOperLine ":" <*> pEqn <*> (pCondeq <|> succeed NoEq)
        <?> (Ceq "error" NoEq undefEqn 0, "conditional equation")
        where ceq n l e conds = Ceq n conds e l

pCondeq = pSym ',' <-*> pKey "if" <-*> pEqs 
          <?> (NoEq, "condition of equation")

pEqs = foldr OneEq NoEq <$> pcurly_psemics pEqn
       <?> (NoEq, "equation list")

pEqn = eqn <$> pExp <*-> pOper "=" <*> pExp
       <?> (undefEqn, "equation")
	   where eqn e1 e2 = Eqn (unlet e1) (unlet e2)

undefEqn = Eqn undefExp undefExp 


